// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit OSThread;

interface

uses
  {$IFDEF WIN32} windows, {$ENDIF} sysutils, classes, types;

type
  TMutexAttribute = (maNormal, maRecursive);
  TMutexLockResult = (NoError, InvalidPriority, MutexAlreadyLocked);

  // This class can be used to establish a critical section. Call the
  // lock method to lock the mutex, and the unlock Method to unlock the
  // mutex. If the mutex is currently locked the thread will be
  // suspended until the mutex becomes unlocked
  TMutex = class
  private
{$IFDEF WIN32}
    FCrit: TRTLCriticalSection; // WIN32 critical section
{$ENDIF}
  public
    constructor Create(ALocked: boolean = false; Attr: TMutexAttribute = maNormal);
    destructor Destroy; override;
    function trylock(): TMutexLockResult;
    procedure lock();
    procedure unlock();
  end;

  // This class acts like a pthreads condition variable. The difference
  // is that this implementation works with the recursive mutex
  // implementation in this library.
  TCondVar = class
  private
{$IFDEF WIN32}
    FWaiters: integer;
    FSema: THandle; // Handle to WIN32 sempahore object.
    FBroadcastEvent: THandle;
    FHaveBroadcast: boolean;
    FWaitersMutex: TMutex;
    //FCrit: TRTLCriticalSection; // WIN32 critical section
{$ENDIF}
    function InternalWait(mutex: TMutex; tmout: Longint): boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function wait(mutex: TMutex): boolean; overload;
    function wait(mutex: TMutex; tmout: Longint): boolean; overload;
    procedure broadcast();
    procedure signal();
  end;

  TSemaphore = class
{$IFDEF WIN32}
    FSema: THandle; // Handle to WIN32 sempahore object.
{$ENDIF}
  public
    constructor Create(val: integer = 0);
    destructor Destroy; override;
    procedure wait();
    procedure post();
  end;

{$DEFINE _THR_CREATE_AND_BLOCK}
  TDetachFlag = (dfDetached, dfNotDetached);
  TThreadErrorType = (teNoError, teInvalidPriority, teMutexAlreadyLocked, teNotStarted, teUnknownError);

{$IFDEF WIN32}
  TThreadKey = DWORD;
{$ENDIF}

  // This class encapsulates the functionality of a thread.
  TORBThread = class
  private
    //FThread: TThread;
    FThreadHandle: THandle;
    FID: THandle;
    FArg: pointer;
    FDetached: TDetachFlag;
    FReturnValue: integer;
    FNo: integer;
{$IFDEF _THR_CREATE_AND_BLOCK}
    FReady: TMutex;
{$ENDIF _THR_CREATE_AND_BLOCK}
    function GetThreadId: integer;
  protected
    function GetNo: integer;
    procedure SetNo(value: integer);
  public
    constructor Create(detached: TDetachFlag = dfNotDetached);
    destructor Destroy; override;
    procedure _thr_startup(arg: pointer);
    // Override this method to provide functionality to your thread.
    procedure _run(arg: pointer); virtual; abstract;
    function start(arg: pointer = nil): TThreadErrorType;
    function create_thread(): TThreadErrorType;
    procedure yield();
    procedure terminate(exitval: pointer);
    procedure priority(new_priority: integer);
    procedure wait();
    // key handling
    class procedure createKey(var key: TThreadKey);
    class procedure deleteKey(key: TThreadKey);
    class function getSpecific(key: TThreadKey): Pointer;
    class procedure setSpecific(key: TThreadKey; value: Pointer);
    
    property ID: integer read GetThreadId;
  end;

  TExecState = (esInit, esActive, esInitShutdown, esShutdown, esTerminated);

  TStateRefCount = class
  private
    FExecState: TExecState;
    FActiveRefs: _ulong;
    FMutex: TMutex;
    FCondVar: TCondVar;
  public
    constructor Create;
    destructor Destroy; override;
    function activeRef(): boolean;
    procedure activeDeref();
    function StateChange(es: TExecState): boolean;
    function State(): TExecState;
    function initShutdown(): boolean;
  end;

  TRWLock = class
  private
    FLocker: TMultiReadExclusiveWriteSynchronizer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure rdlock();
    procedure wrlock();
    procedure rdunlock();
    procedure wrunlock();
  end;

  IAuto_Lock = interface
  end;

  TAuto_Lock = class(TInterfacedObject, IAuto_Lock)
  private
    FMutex: TMutex;
  public
    constructor Create(AMutex: TMutex);
    destructor Destroy; override;
  end;

  TAuto_RDLock = class(TInterfacedObject, IAuto_Lock)
  private
    FRWLock: TRWLock;
  public
    constructor Create(ARWLock: TRWLock);
    destructor Destroy; override;
  end;

  TAuto_WRLock = class(TInterfacedObject, IAuto_Lock)
  private
    FRWLock: TRWLock;
  public
    constructor Create(ARWLock: TRWLock);
    destructor Destroy; override;
  end;

implementation

threadvar
  threadSpecific: pointer;

{ TMutex }

constructor TMutex.Create(ALocked: boolean; Attr: TMutexAttribute);
begin
{ TODO -oOVS : Implement recursive mutex }
{$IFDEF WIN32}
  InitializeCriticalSection(FCrit);
  if ALocked then EnterCriticalSection(FCrit);
{$ENDIF}
end;

destructor TMutex.Destroy;
begin
{$IFDEF WIN32}
  DeleteCriticalSection(FCrit);
{$ENDIF}
  inherited;
end;

// Lock the mutex.
procedure TMutex.lock;
begin
{$IFDEF WIN32}
  EnterCriticalSection(FCrit);
{$ENDIF}
end;

// Try lock the mutex.
function TMutex.trylock: TMutexLockResult;
begin
{$IFDEF WIN32}
  if TryEnterCriticalSection(FCrit) then result := NoError
  else result := MutexAlreadyLocked;
{$ENDIF}
end;

// Unlock the mutex.
procedure TMutex.unlock;
begin
{$IFDEF WIN32}
  LeaveCriticalSection(FCrit);
{$ENDIF}
end;

{ TCondVar }

//
// Wake all waiting threads.
//
procedure TCondVar.broadcast;
{$IFDEF WIN32}
var
  haveW: boolean;
  numW: integer;
{$ENDIF}
begin
{$IFDEF WIN32}
  FWaitersMutex.lock;
  try
    haveW := FWaiters > 0;
    numW := FWaiters;
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  if haveW then begin
    // This is a broadcast event, so set the flag and reset
    // the Event synchronization object.
    FHaveBroadcast := true;
    ResetEvent(FBroadcastEvent);
    // Release the semaphore waiters_ times.
    ReleaseSemaphore(FSema, numW, nil);
    // Wait for each thread to wake.
    // After reset the haveBroadcast_ flag.
    WaitForSingleObject(FBroadcastEvent, INFINITE);
    FHaveBroadcast := false;
  end;
{$ENDIF}
end;

constructor TCondVar.Create;
begin
{$IFDEF WIN32}
  FWaiters := 0;
  FHaveBroadcast := false;;
  FSema := CreateSemaphore(nil, 0, $7fffffff, nil);
  FBroadcastEvent := CreateEvent(nil, true, false, nil);
  Assert((FSema <> INVALID_HANDLE_VALUE) or (FBroadcastEvent <> INVALID_HANDLE_VALUE));
  FWaitersMutex := TMutex.Create;
{$ENDIF}
end;

destructor TCondVar.Destroy;
begin
{$IFDEF WIN32}
  FWaitersMutex.Free;
  CloseHandle(FSema);
  CloseHandle(FBroadcastEvent);
{$ENDIF}
  inherited;
end;

function TCondVar.InternalWait(mutex: TMutex; tmout: Longint): boolean;
var
  //count: integer;
  tm: DWORD;
begin
  result := true;
{$IFDEF WIN32}
  // mutex is locked on entry.
  FWaiters := InterlockedIncrement(FWaiters);
  if tmout < 0 then tm := INFINITE else tm := tmout;
  mutex.unlock();
  try
    if WaitForSingleObject(FSema, tm) = WAIT_ABANDONED then
      raise Exception.Create('System exception');//do not localize
  except
    // Restore the mutex lock count times.
    FWaiters := InterlockedDecrement(FWaiters);
    mutex.lock;
    raise;
  end; { try/except }
  // Restore the mutex lock count times.
  FWaitersMutex.lock;
  try
    Dec(FWaiters);
    if (FWaiters = 0) and FHaveBroadcast then SetEvent(FBroadcastEvent);
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  mutex.lock;
{$ENDIF}
end;

procedure TCondVar.signal;
var
  haveW: boolean;
begin
{$IFDEF WIN32}
  FWaitersMutex.lock;
  try
    haveW := FWaiters > 0;
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  if haveW then ReleaseSemaphore(FSema, 1, nil);
{$ENDIF}
end;

function TCondVar.wait(mutex: TMutex; tmout: Longint): boolean;
begin
  result := InternalWait(mutex, tmout);
end;

function TCondVar.wait(mutex: TMutex): boolean;
begin
  result := InternalWait(mutex, -1);
end;

{ TSemaphore }

constructor TSemaphore.Create(val: integer);
begin
{$IFDEF WIN32}
  FSema := CreateSemaphore(nil, val, $7fffffff, nil);
{$ENDIF}
end;

destructor TSemaphore.Destroy;
begin
{$IFDEF WIN32}
  CloseHandle(FSema);
{$ENDIF}
  inherited;
end;

procedure TSemaphore.post;
begin
{$IFDEF WIN32}
  ReleaseSemaphore(FSema, 1, nil);
{$ENDIF}
end;

procedure TSemaphore.wait;
begin
{$IFDEF WIN32}
  WaitForSingleObject(FSema, INFINITE);
{$ENDIF}
end;

{ TORBThread }

const
  levelMap: array [0..2] of integer = (
{$IFDEF WIN32}
    THREAD_PRIORITY_LOWEST,
    THREAD_PRIORITY_NORMAL,
    THREAD_PRIORITY_HIGHEST
{$ENDIF}
  );

// Start the thread.  
procedure TORBThread._thr_startup(arg: pointer);
begin
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.lock();
{$ENDIF _THR_CREATE_AND_BLOCK}
  _run(arg);
end;

constructor TORBThread.Create(detached: TDetachFlag);
begin
  FDetached := detached;
{$IFDEF _THR_CREATE_AND_BLOCK}
  create_thread();
  FReady := TMutex.Create;
{$ENDIF _THR_CREATE_AND_BLOCK}
end;

function ThreadFunc(thr: TORBThread): integer;
begin
  thr._run(thr.FArg);
  result := thr.FReturnValue;
end;

function TORBThread.create_thread: TThreadErrorType;
begin
{$IFDEF WIN32}
  FThreadHandle := BeginThread(nil, 0, @ThreadFunc, Self, 0, FID);
  if FThreadHandle <> 0 then result := teNoError else result := teUnknownError;
{$ENDIF}
end;

destructor TORBThread.Destroy;
begin
  wait();
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.unlock();
  FReady.Free;
{$ENDIF _THR_CREATE_AND_BLOCK}
  inherited;
end;

// Set the priority of this thread.
procedure TORBThread.priority(new_priority: integer);
begin
{$IFDEF WIN32}
  SetThreadPriority(FThreadHandle, levelMap[new_priority]);
{$ENDIF}
end;

// Start the thread ala JAVA:
function TORBThread.start(arg: pointer): TThreadErrorType;
begin
  FArg := arg;
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.unlock();
  result := teNoError;
{$ELSE}
  result := create_thread();
{$ENDIF _THR_CREATE_AND_BLOCK}
end;

// Tell this thread to die.
procedure TORBThread.terminate(exitval: pointer);
begin
{$IFDEF WIN32}
  ExitThread(PInteger(exitval)^);
{$ENDIF}
end;

// Wait for this thread to finish.
procedure TORBThread.wait;
begin
{$IFDEF WIN32}
  WaitForSingleObject(FThreadHandle, INFINITE);
{$ENDIF}
end;

// Yield the current threads timeslice.
procedure TORBThread.yield;
begin
{$IFDEF WIN32}
  //yield(); //win16
  Sleep(0);
{$ENDIF}
end;

function TORBThread.GetThreadId: integer;
begin
  result := FID;
end;

function TORBThread.GetNo: integer;
begin
  result := FNo;
end;

procedure TORBThread.SetNo(value: integer);
begin
  FNo := value;
end;

class procedure TORBThread.createKey(var key: TThreadKey);
begin

end;

class procedure TORBThread.deleteKey(key: TThreadKey);
begin

end;

class function TORBThread.getSpecific(key: TThreadKey): Pointer;
begin
  result := threadSpecific;
end;

class procedure TORBThread.setSpecific(key: TThreadKey; value: Pointer);
begin
  threadSpecific := value;
end;

{ TRWLock }

constructor TRWLock.Create;
begin
  FLocker := TMultiReadExclusiveWriteSynchronizer.Create;
end;

destructor TRWLock.Destroy;
begin
  FLocker.Free;
  inherited;
end;

procedure TRWLock.rdlock;
begin
  FLocker.BeginRead;
end;

procedure TRWLock.rdunlock;
begin
  FLocker.EndRead;
end;

procedure TRWLock.wrlock;
begin
  FLocker.BeginWrite;
end;

procedure TRWLock.wrunlock;
begin
  FLocker.EndWrite;
end;

{ TAuto_Lock }

constructor TAuto_Lock.Create(AMutex: TMutex);
begin
  FMutex := AMutex;
  FMutex.lock();
end;

destructor TAuto_Lock.Destroy;
begin
  FMutex.unlock();
  inherited;
end;

{ TAuto_RDLock }

constructor TAuto_RDLock.Create(ARWLock: TRWLock);
begin
  FRWLock := ARWLock;
  FRWLock.rdlock();
end;

destructor TAuto_RDLock.Destroy;
begin
  FRWLock.rdunlock();
  inherited;
end;

{ TAuto_WRLock }

constructor TAuto_WRLock.Create(ARWLock: TRWLock);
begin
  FRWLock := ARWLock;
  FRWLock.wrlock();
end;

destructor TAuto_WRLock.Destroy;
begin
  FRWLock.wrunlock();
  inherited;
end;

{ TStateRefCount }

procedure TStateRefCount.activeDeref;
begin
  FMutex.lock();
  try
    FActiveRefs := Pred(FActiveRefs);
    if (FActiveRefs = 0) and (FExecState <> esActive) then
      FCondVar.signal();
  finally
    FMutex.unlock();
  end; { try/finally }
end;

function TStateRefCount.activeRef: boolean;
begin
  FMutex.lock();
  try
    result := FExecState = esActive;
    if result then FActiveRefs := Succ(FActiveRefs);
  finally
    FMutex.unlock();
  end; { try/finally }
end;

constructor TStateRefCount.Create;
begin
  FMutex := TMutex.Create;
  FCondVar := TCondVar.Create;
  FExecState := esInit;
  FActiveRefs := 0;
end;

destructor TStateRefCount.Destroy;
begin
  FCondVar.Free;
  FMutex.Free;
  inherited;
end;

function TStateRefCount.initShutdown: boolean;
begin
  FMutex.lock();
  try
    result := FExecState = esActive;
    if result then FExecState := esInitShutdown;
  finally
    FMutex.unlock();
  end; { try/finally }
end;

function TStateRefCount.State: TExecState;
begin
  result := FExecState;
end;

function TStateRefCount.StateChange(es: TExecState): boolean;
begin
  FMutex.lock();
  try
    result := es > FExecState;
    if result then begin
      FExecState := es;
      // wait until the state change has become effective
      while FActiveRefs <> 0 do
        FCondVar.wait(FMutex);
    end;
  finally
    FMutex.unlock();
  end; { try/finally }
end;

end.
